#ifndef MACROS_H_
#define MACROS_H_

/*E
  LogLevel - Logging level.

  STDOUT (writes to standard output), INFO (writes input parameters,
  properties, boundary conditions, convergence, ...) and DEBUG
  (debug purposes).

  Level: developer
E*/
typedef enum {
  STDOUT, INFO, DEBUG
} LogLevel;

/*M
  WEST - Macro for identifying the computational molecule:
  west neighbour node.

  Level: advanced
M*/
#define WEST 0
/*M
  EAST - Macro for identifying the computational molecule:
  east neighbour node.

  Level: advanced
M*/
#define EAST 1
/*M
  SOUTH - Macro for identifying the computational molecule:
  south neighbour node.

  Level: advanced
M*/
#define SOUTH 2
/*M
  NORTH - Macro for identifying the computational molecule:
  north neighbour node.

  Level: advanced
M*/
#define NORTH 3
/*M
  LOW - Macro for identifying the computational molecule:
  low neighbour node.

  Level: advanced
M*/
#define LOW 4
/*M
  HIGH - Macro for identifying the computational molecule:
  high neighbour node.

  Level: advanced
M*/
#define HIGH 5
/*M
  NODEP - Macro for identifying the computational molecule:
  node itself.

  Level: advanced
M*/
#define NODEP (2 * Input.ndim)

/*M
  XDIR - Macro for axis: X direction.

  Level: advanced
M*/
#define XDIR 0
/*M
  YDIR - Macro for axis: Y direction.

  Level: advanced
M*/
#define YDIR 1
/*M
  ZDIR - Macro for axis: Z direction.

  Level: advanced
M*/
#define ZDIR 2

/*M
  GREAT - Constant for a great double value (1e08).

  Level: advanced
M*/
#define GREAT 1.0E8
/*M
  TINY - Constant for a tiny double value (1e-15).

  Level: advanced
M*/
#define TINY 1.0E-15

/*M
  L - Grid length (X direction).

  Level: advanced
M*/
#define L (Grid.Xnode[XDIR][Grid.Nnode[XDIR] - 1])
/*M
  H - Grid height (Y direction).

  Level: advanced
M*/
#define H (Grid.Xnode[YDIR][Grid.Nnode[YDIR] - 1] )
/*M
  REYNOLDS - Reynolds number.

  Option database keys:
+ \-rho - RHO value.
. \-uref - Constant value por the U direction inlet.
- \-mu - MU value.

  Level: advanced
M*/
#define REYNOLDS (Input.rho * H * Input.uref / Input.mu)

/*M
  MIN3 - Macro for the minimum value among three values

  Input parameters:
. a, b, c - Values to get the minimum.

  Level: intermediate
M*/
#define MIN3(a, b, c) (MIN(MIN(a, b), c))

/*E
  TransientOrder - Transient order.

  Values are: ZERO (for steady flows), EUL1 (first order Euler), EUL2
  (second order Euler), ADM2 (second order Adams-Moulton), ADM3
  (third order Adams-Moulton).

  Level: advanced
E*/
typedef enum {
  ZERO, EUL1, EUL2, ADM2, ADM3
} TransientOrder;

/*E
  ktypeProp - Property type.

  Defines if the property depends on time (TRAN), the variable values
  (VARI), the geometry (GEOM) or nothing (CONS).

  Level: advanced
E*/
typedef enum {
  CONS, GEOM, VARI, TRAN
} ktypeProp;

/*E
  StorageType - Storage type.

  Defines if the property should be stored as a single real number (SCALAR)
  or an array of real numbers (ARRAY).

  Level: advanced
E*/
typedef enum {
  SCALAR, ARRAY
} StorageType;

/*E
  ktypeGrid - Grid type.

  Values are: RECT (rectangular grids) and CYLI (cylindrical grids).

  Level: advanced
E*/
typedef enum {
  RECT, CYLI
} ktypeGrid;

/*E
  ktypeIp - Interpolation type for the values in the cell faces.

  Values can be: NIP, UPWIND, CDS, HYBRID, QUIKC, CUI, SMART, MUSCL,
  NOTABLE and VAN_LEER. TODO document me!! (explain values)

  Level: advanced
E*/
typedef enum {
  NIP, UPWIND, CDS, HYBRID, QUICK, CUI, SMART, MUSCL,
  NOTABLE, VAN_LEER
} ktypeIp;

/*E
  ktypeIp - Boundary condition type.

  Boundary condition can be: wall (WALL), moving wall (MOVING_WALL),
  inlet (INLET), outlet (OUTFLOW), newman outlet pressure
  (OUTPRESS_NEWMAN), extrapolation outlet pressure (OUTPRESS_EXTRAP),
  outconv? (OUTCONV), symmetry (SYMMETRY) and a fix value (FIX_VALUE).
  TODO document me!! (outconv)

  Level: advanced
E*/
typedef enum {
  NONE, WALL, MOVING_WALL, INLET, OUTFLOW, OUTPRESS_NEWMAN,
  OUTPRESS_EXTRAP, OUTCONV, SYMMETRY, FIX_VALUE
} BcondType;

/*E
  contEqType - Type of continuity equation.

  Option database keys:
. \-cont_eq - Continuity equation solving method

  Level: advanced
E*/
typedef enum {
  POISSON, SIMPLE, SIMPLEC, SIMPLER
} contEqType;

/*E
  InletFunction - Function for the inlet boundary conditions.

  Option database keys:
. \-inlet - Inlet function

  Level: advanced
E*/
typedef enum {
  INLET_U, INLET_F, INLET_RHO
} InletFunction;

/*E
  RestartType - Restarting option.

  Values are: RESTART_NONE (no results restarted), RESTART_VARIABLES 
  (restart only values for the unknowns), RESTART_INSTANT_MEANS 
  (restart the instant means, but not the variance means), 
  RESTART_VARIANCE_MEANS (restart all the values).

  Level: advanced
E*/
typedef enum {
  RESTART_NONE, RESTART_VARIABLES, RESTART_INSTANT_MEANS, 
  RESTART_VARIANCE_MEANS
} RestartType;

/*M
  U - Index of the U variable (X direction velocity).

  Level: advanced
M*/
#define U (varU->index)
/*M
  V - Index of the V variable (Y direction velocity).

  Level: advanced
M*/
#define V (varV->index)
/*M
  W - Index of the W variable (Z direction velocity).

  Level: advanced
M*/
#define W (varW->index)
/*M
  P - Index of the P variable (pressure).

  Level: advanced
M*/

#define P (varP->index)
/*M
  T - Index of the T variable (temperature).

  Level: advanced
M*/

#define T (varT->index)
/*M
  KE - Index of the KE variable TODO document me!!

  Level: advanced
M*/
#define KE (varKE->index)
/*M
  EP - Index of the EP variable TODO document me!!

  Level: advanced
M*/
#define EP (varEP->index)

/*M
  U_NAME - Name of the U variable (X direction velocity).

  Level: advanced
M*/
#define U_NAME "U"
/*M
  V_NAME - Name of the V variable (Y direction velocity).

  Level: advanced
M*/
#define V_NAME "V"
/*M
  W_NAME - Name of the W variable (Z direction velocity).

  Level: advanced
M*/
#define W_NAME "W"
/*M
  P_NAME - Name of the P variable (pressure).

  Level: advanced
M*/
#define P_NAME "P"
/*M
  T_NAME - Name of the T variable (temperature).

  Level: advanced
M*/
#define T_NAME "T"
/*M
  KE_NAME - Name of the KE variable TODO document me!!

  Level: advanced
M*/
#define KE_NAME "KE"
/*M
  EP_NAME - Name of the EP variable TODO document me!!

  Level: advanced
M*/
#define EP_NAME "EP"

/*M
  LOCAL_CELL - Calculates the index of a cell in a processor (the local
  index) from its coordinates.

  Input parameters:
. i, j, k - Global coordinates of the cell.

  Level: advanced
M*/
#define LOCAL_CELL(i, j, k) \
  ((k - Local.ghostMeshRegion.lowerCorner[ZDIR]) * \
      Local.ghostMeshRegion.nCells[XDIR] * \
      Local.ghostMeshRegion.nCells[YDIR] + \
  (j - Local.ghostMeshRegion.lowerCorner[YDIR]) * \
      Local.ghostMeshRegion.nCells[XDIR] + \
  (i - Local.ghostMeshRegion.lowerCorner[XDIR]))

/*M
  PATCH_CELL - Calculates the index of a cell in a patch
  from its coordinates.

  Input parameters:
+ i, j, k - Global coordinates of the cell.
- patch - Patch to calculate the index.

  Level: advanced
M*/
#define PATCH_CELL(i, j, k, patch, info) \
  ((k - (patch->z + info.gzs)) * info.gxm * info.gym + \
   (j - (patch->y + info.gys)) * info.gxm + \
   (i - (patch->x + info.gxs)))

/*M
  BEGIN_LOCAL_LOOP - Begins a loop over the local cells of the given patch.

  Input parameters:
+ patch - The patch with the cells to iterate.
. info - DALocalInfo object associated with the patch's DA.
- I - 3D iterator.

  Level: advanced
M*/
#define BEGIN_LOCAL_LOOP(patch, info, I) \
for(I[ZDIR] = patch->z + info.zs; I[ZDIR] < patch->z + info.zs + info.zm; I[ZDIR]++) \
for(I[YDIR] = patch->y + info.ys; I[YDIR] < patch->y + info.ys + info.ym; I[YDIR]++) \
for(I[XDIR] = patch->x + info.xs; I[XDIR] < patch->x + info.xs + info.xm; I[XDIR]++)

/*M
  IDIR - Determines the direction of the neighbor.

  Input parameters:
. ineighb - Neighbor index.

  Level: advanced

.seealso WEST, EAST, NORTH, SOUTH, LOW, HIGH
M*/
#define IDIR(ineighb) (ineighb / 2)
/*M
  STEP - Determines the "step" of the neighbor, i.e if the neighbor is
  forwards (step > 0) or backwards (step < 0).

  Input parameters:
. ineighb - Neighbor index.

  Level: advanced

.seealso WEST, EAST, NORTH, SOUTH, LOW, HIGH
M*/
#define STEP(ineighb) (2 * (ineighb % 2) - 1)

/*M
  LOCAL_CELL_NEIGHB_EXISTS - Determines if the specified neighbour is
  in the local subdomain.

  Input parameters:
+ I - Index of the local cell.
. idir - Index direction of the neighbor to evaluate.
- step - Step of the neighbor to evaluate.

  Level: advanced

.seealso IDIR, STEP
M*/
#define LOCAL_CELL_NEIGHB_EXISTS(I, idir, step) \
  (I[idir] + step <= Grid.Nnode[idir] - 1 && I[idir] + step >= 0)

/*M
  PTR_SYS - Gets the system (coupled or segregated) associated with the
  given variable.

  Input parameters:
. ivar - Variable index.

  Level: advanced
M*/
#define PTR_SYS(ivar) ((ivar < Input.ncoupvar) ? \
    &CoupledSys : &SegregatedSys)

/*M
  ROW - The row of the system corresponding to the specified cell and
  equation.

  Input parameters:
+ Sys - Pointer to the evaluating system.
. ieq - Equation index.
- icell - Local cell index.

  Level: advanced
M*/
#define ROW(Sys, ieq, icell) (Sys->ltog[icell * Sys->ndof \
                                        + (ieq - Sys->ivar)])

/*M
  LINEAR_BACKW - Gets the value of the linear backwards interpolation
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define LINEAR_BACKW(idir, icell) (Interp.linear[idir][icell])
/*M
  UPWIND_BACKW - Gets the value of the upwind backwards interpolation
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define UPWIND_BACKW(idir, icell) (Interp.upwind[idir][icell])
/*M
  ICELL_UPW_BACKW - Gets the index of the upwind backwards cell
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_UPW_BACKW(idir, icell) (Interp.icell_upw[idir][icell])
/*M
  ICELL_UUPW_BACKW - Gets the index of the up-upwind backwards cell
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_UUPW_BACKW(idir, icell) (Interp.icell_uupw[idir][icell])
/*M
  ICELL_DOW_BACKW - Gets the index of the downwind backwards cell
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_DOW_BACKW(idir, icell) (Interp.icell_dow[idir][icell])

/*M
  LINEAR_FORW - Gets the value of the linear forwards interpolation
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define LINEAR_FORW(idir, icell) \
  (1 - Interp.linear[idir][Local.icell_neighb[2 * idir + 1][icell]])
/*M
  UPWIND_FORW - Gets the value of the upwind forwards interpolation
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define UPWIND_FORW(idir, icell) \
  (1 - Interp.upwind[idir][Local.icell_neighb[2 * idir + 1][icell]])
/*M
  ICELL_UPW_FORW - Gets the index of the upwind forwards cell
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_UPW_FORW(idir, icell) \
  (Interp.icell_upw[idir][Local.icell_neighb[2 * idir + 1][icell]])
/*M
  ICELL_UUPW_FORW - Gets the index of the up-upwind forwards cell
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_UUPW_FORW(idir, icell) \
  (Interp.icell_uupw[idir][Local.icell_neighb[2 * idir + 1][icell]])
/*M
  ICELL_DOW_FORW - Gets the index of the downwind forwards cell
  for the given cell and direction.

  Input parameters:
+ idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_DOW_FORW(idir, icell) \
  (Interp.icell_dow[idir][Local.icell_neighb[2 * idir + 1][icell]])

/*M
  LINEAR - Gets the value of the linear interpolation
  for the given cell and neighbor.

  Input parameters:
+ ineighb - Neighbor index.
- icell - Cell index.

  Level: advanced
M*/
#define LINEAR(ineighb, icell) \
  (STEP(ineighb) < 0 ? LINEAR_BACKW(IDIR(ineighb), icell) : \
      LINEAR_FORW(IDIR(ineighb), icell))
/*M
  ICELL_UPW - Gets the index of the upwind cell
  for the given cell and neighbor.

  Input parameters:
+ ineighb - Neighbor index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_UPW(ineighb, icell) \
  (STEP(ineighb) < 0 ? ICELL_UPW_BACKW(IDIR(ineighb), icell) : \
      ICELL_UPW_FORW(IDIR(ineighb), icell))
/*M
  ICELL_UUPW - Gets the index of the up-upwind cell
  for the given cell and neighbor.

  Input parameters:
+ ineighb - Neighbor index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_UUPW(ineighb, icell) \
  (STEP(ineighb) < 0 ? ICELL_UUPW_BACKW(IDIR(ineighb), icell) : \
      ICELL_UUPW_FORW(IDIR(ineighb), icell))
/*M
  ICELL_DOW - Gets the index of the downwind cell
  for the given cell and neighbor.

  Input parameters:
+ ineighb - Neighbor index.
- icell - Cell index.

  Level: advanced
M*/
#define ICELL_DOW(ineighb, icell) \
  (STEP(ineighb) < 0 ? ICELL_DOW_BACKW(IDIR(ineighb), icell) : \
      ICELL_DOW_FORW(IDIR(ineighb), icell))

/*M
  VALUE - Gets the value of the variable in the given cell.

  Input parameters:
+ var - Pointer to the variable to evaluate.
- icell - Cell index.

  Level: advanced
M*/
#define VALUE(var, icell) (var->Array[icell])
/*M
  VALUE_FACE_IVAR_UPW_BACKW - Gets the value of the variable in the
  upwind backwards neighbor in the given direction.

  Input parameters:
+ var - Pointer to the variable to evaluate.
. idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define VALUE_FACE_IVAR_UPW_BACKW(var, idir, icell) \
  (VALUE(var, ICELL_UPW_BACKW(idir, icell)))
/*M
  VALUE_FACE_IVAR_UPW_FORW - Gets the value of the variable in the
  upwind forwards neighbor in the given direction.

  Input parameters:
+ var - Pointer to the variable to evaluate.
. idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define VALUE_FACE_IVAR_UPW_FORW(var, idir, icell) \
  (VALUE(var, ICELL_UPW_FORW(idir, icell)))
/*M
  VALUE_FACE_IVAR_LINEAR - Gets the value of the variable in the
  face (linear interpolation) between the given cell and the
  specified neighbor.

  Input parameters:
+ var - Pointer to the variable to evaluate.
. ineighb - Neighbor index.
- icell - Cell index.

  Level: advanced
M*/
#define VALUE_FACE_IVAR_LINEAR(var, ineighb, icell) \
  (LINEAR(ineighb,icell) * VALUE(var, Local.icell_neighb[ineighb][icell]) + \
      (1 - LINEAR(ineighb, icell)) * VALUE(var, icell))

/*M
  VALUE_PROP - Gets the value of a property (constant or not) in a cell.

  Input parameters:
+ prop - Property to evaluate.
- icell - Cell index.

  Level: advanced
M*/
#define VALUE_PROP(prop, icell) \
  (prop->ktype == CONS ? prop->value : prop->array[icell])

/*M
  CONV_BACKW - Gets the value of the Conv property in the given cell
  with the porosity backwards in the specified direction.

  Level: advanced

.seealso Conv
M*/
#define CONV_BACKW(idir, icell) (-PROP_BACKW(Conv, idir, icell))
/*M
  CONV_FORW - Gets the value of the Conv property in the given cell
  with the porosity forwards in the specified direction.

  Level: advanced

.seealso Conv
M*/
#define CONV_FORW(idir,icell) PROP_FORW(Conv, idir, icell)

/*M
  DIFF_BACKW - Gets the value of the Diff property in the given cell
  with the porosity backwards in the specified direction.

  Level: advanced

.seealso Diff
M*/
#define DIFF_BACKW(idir,icell) PROP_BACKW(Diff, idir, icell)
/*M
  DIFF_FORW - Gets the value of the Diff property in the given cell
  with the porosity forwards in the specified direction.

  Level: advanced

.seealso Diff
M*/
#define DIFF_FORW(idir,icell) PROP_FORW(Diff, idir, icell)
/*M
  MICORR_BACKW - Gets the value of the MICorr property in the given cell
  with the porosity backwards in the specified direction.

  Level: advanced

.seealso MICorr
M*/
#define MICORR_BACKW(idir,icell) PROP_BACKW(MICorr, idir, icell)
/*M
  MICORR_FORW - Gets the value of the MICorr property in the given cell
  with the porosity forwards in the specified direction.

  Level: advanced

.seealso MICorr
M*/
#define MICORR_FORW(idir,icell) PROP_FORW(MICorr, idir, icell)
/*M
  MIPRESCONT_BACKW - Gets the value of the MIPresCont property in the
  given cell with the porosity backwards in the specified direction.

  Level: advanced

.seealso MIPresCont
M*/
#define MIPRESCONT_BACKW(idir,icell) PROP_BACKW(MIPresCont, idir, icell)
/*M
  MIPRESCONT_FORW - Gets the value of the MIPresCont property in the
  given cell with the porosity forwards in the specified direction.

  Level: advanced

.seealso MIPresCont
M*/
#define MIPRESCONT_FORW(idir,icell) PROP_FORW(MIPresCont, idir, icell)

/*M
  PROP_BACKW - Gets the value of a property in the given cell with the
  porosity backwards in the specified direction.

  Input parameters:
+ prop - Property to evaluate.
. idir - Direction index.
- icell - Cell index.

  Level: advanced.
M*/
#define PROP_BACKW(prop, idir, icell) \
  (prop->array[Input.ndim * icell + idir] * Local.porosity[2 * idir][icell])
/*M
  PROP_FORW - Gets the value of a property in the given cell with the
  porosity forwards in the specified direction. NOTE: Use ONLY with
  with properties with the same degrees of freedom as dimensions in
  the problem (dof == Input.ndiom).

  Input parameters:
+ prop - Property to evaluate.
. idir - Direction index.
- icell - Cell index.

  Level: advanced.
M*/
#define PROP_FORW(prop, idir, icell) \
  (prop->array[Input.ndim * Local.icell_neighb[2 * idir + 1][icell] + idir] * \
      Local.porosity[2 * idir + 1][icell])

/*M
  COEF_TRANS_IMP - Gets the transient coefficient for the given
  transient order.

  Input parameters:
. norder - Transient order.

  Level: advanced

.seealso TransientOrder
M*/
#define COEF_TRANS_IMP(norder) (norder == EUL1 ? (1.0) : \
    norder == EUL2 ? (1.5) : \
    norder == ADM2 ? (2.0) : (2.4))
/*M
  COEF_TRANS_EXP1 - TODO document me!!

  Input parameters:
. norder - Transient order.

  Level: advanced

.seealso TransientOrder
M*/
#define COEF_TRANS_EXP1(norder) (norder == EUL2 ? (2) : \
    COEF_TRANS_IMP(norder))
/*M
  COEF_TRANS_EXP2 - TODO document me!!

  Input parameters:
. norder - Transient order.

  Level: advanced

.seealso TransientOrder
M*/
#define COEF_TRANS_EXP2(norder) (norder == EUL2 ? (-0.5) : 0)
/*M
  ALPHA_TRANS_MI - TODO document me!!

  Input parameters:
+ coef_trans -
. idir - Direction index.
- icell - Cell index.

  Level: advanced
M*/
#define ALPHA_TRANS_MI(coef_trans, idir, icell) \
  (Interp.linear[idir][icell] * VALUE_PROP(Rho, \
  Local.icell_neighb[2 * idir][icell]) * \
  Geom.Volume[Local.icell_neighb[2 * idir][icell]] * \
  VALUE_PROP(Omica, Local.icell_neighb[2 * idir][icell]) * \
  coef_trans / Input.deltat + (1 - Interp.linear[idir][icell]) * \
  VALUE_PROP(Rho, icell) * Geom.Volume[icell] * \
  VALUE_PROP(Omica, icell) * coef_trans / Input.deltat)

#define PropGetArray(prop) \
  ((prop != PETSC_NULL && prop->storage == ARRAY) ? \
      VecGetArray(prop->values, &prop->array) : 0)

#define PropRestoreArray(prop) \
  ((prop != PETSC_NULL && prop->storage == ARRAY) ? \
      VecRestoreArray(prop->values, &prop->array) : 0)

#endif /* MACROS_H_ */
